import math, random

import gym
import numpy as np

import sys
import torch
import torch.nn as nn
import torch.optim as optim
import torch.autograd as autograd 
import torch.nn.functional as F

import time

from utils.wrappers import make_env

import ale_py

import matplotlib.pyplot as plt

from collections import deque

import argparse

import sys
sys.path.append("..")
sys.path.append("../..")

import general_energy_nets.predictive_coding as pc

ENV_ID = "PongNoFrameskip-v4"
NUM_SEEDS = 1
RAM = False

EPSILON_START = 1.0
EPSILON_FINAL = 0.01
EPSILON_DECAY = 15000

STEP_FREQ = 1
MAX_EPISODE_LENGTH = 10000
RANDOM_FRAMES = 0

MAX_FRAMES = 1400000
BATCH_SIZE = 32
GAMMA = 0.99

OPTIM = 'Adam'
LR = 0.0001
SYNC_FREQ = 1000

WARMUP_FRAMES = 10000
BUFFER_SIZE = 100000

REWARD_EVERY_N = 10 # 10 for Pong 100 for breakout

class ReplayBuffer(object):
    def __init__(self, capacity):
        self.buffer = deque(maxlen=capacity)
    
    def push(self, state, action, reward, next_state, done):
        state      = np.expand_dims(state, 0)
        next_state = np.expand_dims(next_state, 0)
            
        self.buffer.append((state, action, reward, next_state, done))
    
    def sample(self, batch_size):
        state, action, reward, next_state, done = zip(*random.sample(self.buffer, batch_size))
        return np.concatenate(state), action, reward, np.concatenate(next_state), done
    
    def __len__(self):
        return len(self.buffer)

class CnnPCDQN(nn.Module):
    def __init__(self, input_shape, num_actions):
        super(CnnPCDQN, self).__init__()
        
        self.conv = nn.Sequential(
            nn.Conv2d(input_shape[0], 32, kernel_size=8, stride=4),
            pc.PCLayer(),
            nn.ReLU(),
            nn.Conv2d(32, 64, kernel_size=4, stride=2),
            pc.PCLayer(),
            nn.ReLU(),
            nn.Conv2d(64, 64, kernel_size=3, stride=1),
            pc.PCLayer(),
            nn.ReLU()
        )
        
        self.mlp = nn.Sequential(
            nn.Linear(self.conv_out_size(input_shape), 512),
            pc.PCLayer(),
            nn.ReLU(),
            nn.Linear(512, num_actions)
        )
        
    def conv_out_size(self, shape):
        o = self.conv(torch.zeros(1, *shape))
        return int(np.prod(o.size()))

    def forward(self, x):
        conv_out = self.conv(x).view(x.size()[0], -1)
        return self.mlp(conv_out)
    
    @torch.no_grad()
    def act(self, state, epsilon):
        if random.random() > epsilon:
            state   = torch.FloatTensor(np.float32(state)).unsqueeze(0).to(device)
            q_value = self.forward(state)
            action  = q_value.max(1)[1].data[0].item()
        else:
            action = random.randrange(env.action_space.n)
        return action

class PCDQN(nn.Module):
    def __init__(self, input_shape, num_actions):
        super(PCDQN, self).__init__()
        
        self.layers = nn.Sequential(
            nn.Linear(input_shape[0], 128),
            pc.PCLayer(),
            nn.ReLU(),
            nn.Linear(128, 128),
            pc.PCLayer(),
            nn.ReLU(),
            nn.Linear(128, num_actions)
        )
        
    def forward(self, x):
        x = x.view(x.size(0),-1)
        fx = x.float() / 256.0
        return self.layers(fx)
    
    @torch.no_grad()
    def act(self, state, epsilon):
        if random.random() > epsilon:
            state   = torch.FloatTensor(state).unsqueeze(0).to(device)
            q_value = self.forward(state)
            action  = q_value.max(1)[1].data[0].item()
        else:
            action = random.randrange(env.action_space.n)
        return action

def update_target(current_model, target_model):
    parameters = current_model.state_dict()
    if RAM:
        try:
            del parameters['layers.1._x']
            del parameters['layers.4._x']
        except:
            pass
    else:
        try:
            del parameters['conv.1._x']
            del parameters['conv.4._x']
            del parameters['conv.7._x']
            del parameters['mlp.1._x']
        except:
            pass
    target_model.load_state_dict(parameters)

def loss_fn(output, _expected_q_value, _action):
    q_value = output.gather(1, _action.unsqueeze(1)).squeeze(1) 
    return (q_value - _expected_q_value).pow(2).sum() * 0.5

def compute_td_loss_without_clamp(batch_size):
    state, action, reward, next_state, done = replay_buffer.sample(batch_size)

    state      = torch.FloatTensor(np.float32(state)).to(device)
    next_state = torch.FloatTensor(np.float32(next_state)).to(device)
    action     = torch.LongTensor(action).to(device)
    reward     = torch.FloatTensor(reward).to(device)
    done       = torch.BoolTensor(done).to(device)

    #q_values      = current_model(state)
    #q_value       = q_values.gather(1, action.unsqueeze(1)).squeeze(1) 
    
    current_model.eval()
    
    with torch.no_grad():
        next_q_values = current_model(next_state)
        next_q_state_values = target_model(next_state)
        next_q_value = next_q_state_values.gather(1, torch.max(next_q_values, 1)[1].unsqueeze(1)).squeeze(1)
        next_q_value[done] = 0.0
        next_q_value = next_q_value.detach()

    expected_q_value = reward + GAMMA * next_q_value
    
    current_model.train()
    results = pc_trainer.train_on_batch(
        inputs=state,
        loss_fn=loss_fn,
        loss_fn_kwargs={
            '_expected_q_value': expected_q_value,
            '_action': action
        },
        is_log_progress=False,
        is_return_results_every_t=False,
        is_checking_after_callback_after_t=False,
    )
    
    return results

def plot(reward_matrix, confidence_intervals=False):
    plt.style.use('ggplot')
    average_reward = np.mean(np.mean(reward_matrix, axis=0))
    plt.figure(figsize=(10, 5))
    plt.title('{} : DQN with PC (Avg Reward = {:.2f})'.format(ENV_ID, average_reward))
    reward_matrix_mean = np.mean(reward_matrix, axis=0)
    plt.plot([x * REWARD_EVERY_N for x in range(reward_matrix.shape[1])], reward_matrix_mean, linewidth=2, color='firebrick', alpha=0.8)

    if confidence_intervals:
        upper_confidence_interval = []
        lower_confidence_interval = []

        for episode in reward_matrix.transpose():
            bootstrapped_means = []
            for B in range(2000):
                bootstrapped_sample = np.random.choice(episode, len(episode))
                bootstrapped_means.append(np.mean(bootstrapped_sample))
            bootstrapped_percentiles = np.percentile(np.array(bootstrapped_means), [2.5, 97.5])
            upper_confidence_interval.append(bootstrapped_percentiles[1])
            lower_confidence_interval.append(bootstrapped_percentiles[0])

        plt.fill_between([x * REWARD_EVERY_N for x in range(reward_matrix.shape[1])], upper_confidence_interval, lower_confidence_interval, color='firebrick', alpha=0.4)

    plt.xlabel('frames')
    plt.ylabel('test reward')
    plt.savefig('./imgs/DQN_pc_{}_seeds={}_optim={}_lr={}.png'.format(ENV_ID, NUM_SEEDS, OPTIM, LR))
    plt.close('all')

def save(reward_matrix):
    with open('./data/DQN_pc_{}_seeds={}_optim={}_lr={}.npy'.format(ENV_ID, NUM_SEEDS, OPTIM, LR), 'wb') as outfile:
        np.save(outfile, reward_matrix)

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--env-id", default=ENV_ID)
    parser.add_argument("--n-seed", default=NUM_SEEDS)
    parser.add_argument("--max-frames", default=MAX_FRAMES)
    parser.add_argument("--epsilon-decay", default=EPSILON_DECAY)
    parser.add_argument("--epsilon-final", default=EPSILON_FINAL)
    parser.add_argument("--random-frames", default=RANDOM_FRAMES)
    parser.add_argument("--batch-size", default=BATCH_SIZE)
    parser.add_argument("--optim", default=OPTIM)
    parser.add_argument("--lr", default=LR)
    parser.add_argument("--step-freq", default=STEP_FREQ)
    parser.add_argument("--sync-freq", default=SYNC_FREQ)
    parser.add_argument("--warmup-frames", default=WARMUP_FRAMES)
    parser.add_argument("--buffer-size", default=BUFFER_SIZE)
    parser.add_argument("--reward-every", default=REWARD_EVERY_N)
    args = parser.parse_args()

    ENV_ID = args.env_id
    NUM_SEEDS = int(args.n_seed)
    MAX_FRAMES = int(args.max_frames)
    RAM = False if ENV_ID.find("-ram") == -1 else True
    EPSILON_DECAY = int(args.epsilon_decay)
    EPSILON_FINAL = float(args.epsilon_final)
    RANDOM_FRAMES = int(args.random_frames)
    BATCH_SIZE = int(args.batch_size)
    OPTIM = args.optim
    LR = float(args.lr)
    STEP_FREQ = int(args.step_freq)
    SYNC_FREQ = int(args.sync_freq)
    WARMUP_FRAMES = int(args.warmup_frames)
    BUFFER_SIZE = int(args.buffer_size)
    REWARD_EVERY_N = int(args.reward_every)

    assert torch.cuda.is_available()
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu" )

    reward_matrix = []

    for SEED in range(1, NUM_SEEDS+1):
        env = make_env(ENV_ID, ram=RAM)

        torch.manual_seed(SEED)
        np.random.seed(SEED)
        random.seed(SEED)
        env.action_space.seed(SEED)
        env.seed(SEED)

        if RAM:
            current_model = PCDQN(env.observation_space.shape, env.action_space.n).to(device)
            target_model  = PCDQN(env.observation_space.shape, env.action_space.n).to(device)
        else:
            current_model = CnnPCDQN(env.observation_space.shape, env.action_space.n).to(device)
            target_model  = CnnPCDQN(env.observation_space.shape, env.action_space.n).to(device)

        target_model.eval()

        update_target(current_model, target_model)

        if OPTIM == 'Adam':
            optimizer = optim.Adam
        elif OPTIM == 'SGD':
            optimizer = optim.SGD
        else:
            raise ValueError

        pc_trainer = pc.PCTrainer(
            current_model,
            T=32,
            update_x_at='all',
            optimizer_x_fn=optim.SGD,
            optimizer_x_kwargs={"lr": 0.05},
            x_lr_discount=0.5,
            update_p_at='last',
            optimizer_p_fn=optimizer,
            optimizer_p_kwargs={"lr": LR},
            plot_progress_at=[],
        )

        replay_buffer = ReplayBuffer(BUFFER_SIZE)

        all_mean_rewards = []

        losses = []
        all_rewards = []
        episode_reward = 0
        episode_frames = 0
        episode_idx = 0

        state = env.reset()
        for frame_idx in range(1, MAX_FRAMES + 1):
            epsilon = 1.0 if frame_idx < RANDOM_FRAMES else EPSILON_FINAL + (EPSILON_START - EPSILON_FINAL) * np.exp(-1. * (frame_idx - RANDOM_FRAMES) / EPSILON_DECAY)

            current_model.eval()
            action = current_model.act(state, epsilon)
    
            next_state, reward, done, _ = env.step(action)
            replay_buffer.push(state, action, reward, next_state, done)
            
            state = next_state
            episode_reward += reward
            episode_frames += 1

            if done or episode_frames >= MAX_EPISODE_LENGTH:
                state = env.reset()
                all_rewards.append(episode_reward)
                episode_reward = 0
                episode_frames = 0
                episode_idx += 1
                if episode_idx % REWARD_EVERY_N == 0:
                    all_mean_rewards.append(np.mean(all_rewards[-REWARD_EVERY_N:]))
                    print("reward %.3f  frame %d" % (all_mean_rewards[-1], frame_idx))

            if (len(replay_buffer) > WARMUP_FRAMES) and (frame_idx % STEP_FREQ == 0):
                loss = compute_td_loss_without_clamp(BATCH_SIZE)['loss'][0]
                losses.append(loss)
                    
            if frame_idx % SYNC_FREQ == 0:
                update_target(current_model, target_model)

        reward_matrix.append(all_mean_rewards)

    reward_matrix = np.array(reward_matrix)
    plot(reward_matrix, confidence_intervals=(NUM_SEEDS >= 3))
    save(reward_matrix)