import math, random

import gym
import numpy as np

import sys
import torch
import torch.nn as nn
import torch.optim as optim
import torch.autograd as autograd 
import torch.nn.functional as F

import sys
sys.path.append("..")

import general_energy_nets.predictive_coding as pc

import matplotlib.pyplot as plt

from collections import deque

import argparse

ENV_ID = "CartPole-v1"
NUM_SEEDS = 10

EPSILON_START = 1.0
EPSILON_FINAL = 0.01
EPSILON_DECAY = 500

NUM_EPISODES = 100 #100
BATCH_SIZE = 64
GAMMA = 0.98

#ACTF_FN = 'ReLU'
PC_LR = 0.001
PC_OPTIM = 'Adam'

BP_LR = 0.001
BP_OPTIM = 'Adam'

#BIG_T = 128
#X_LR = 0.01
#ACTF_FN = 'ReLU'
#UPDATE_P = 'last'
#P_OPTIM = 'Adam'
#CLAMP = False
#USE_PC_PLOTTER = False

class ReplayBuffer(object):
    def __init__(self, capacity):
        self.buffer = deque(maxlen=capacity)
    
    def push(self, state, action, reward, next_state, done):
        state      = np.expand_dims(state, 0)
        next_state = np.expand_dims(next_state, 0)
            
        self.buffer.append((state, action, reward, next_state, done))
    
    def sample(self, batch_size):
        state, action, reward, next_state, done = zip(*random.sample(self.buffer, batch_size))
        return np.concatenate(state), action, reward, np.concatenate(next_state), done
    
    def __len__(self):
        return len(self.buffer)

def loss_fn(output, _expected_q_value, _action):
    q_value = output.gather(1, _action.unsqueeze(1)).squeeze(1) 
    return (q_value - _expected_q_value).pow(2).sum() * 0.5

def clamp_loss_fn(output, _targets):
    return (output - _targets).pow(2).sum() * 0.5

def update_target(current_model, target_model):
    parameters = current_model.state_dict()
    try:
        del parameters['layers.1._x']
        del parameters['layers.4._x']
    except:
        pass
    target_model.load_state_dict(parameters)

#target_model.load_state_dict(current_model.state_dict())

def compute_td_loss_without_clamp(batch_size):
    state, action, reward, next_state, done = replay_buffer.sample(batch_size)

    state      = torch.FloatTensor(np.float32(state)).to(device)
    next_state = torch.FloatTensor(np.float32(next_state)).to(device)
    action     = torch.LongTensor(action).to(device)
    reward     = torch.FloatTensor(reward).to(device)
    done       = torch.BoolTensor(done).to(device)

    #q_values      = current_model(state)
    #q_value       = q_values.gather(1, action.unsqueeze(1)).squeeze(1) 
    
    current_model.eval()
    
    with torch.no_grad():
        next_q_values = current_model(next_state)
        next_q_state_values = target_model(next_state) 
        next_q_value = next_q_state_values.gather(1, torch.max(next_q_values, 1)[1].unsqueeze(1)).squeeze(1)
        next_q_value[done] = 0.0
        next_q_value = next_q_value.detach()

    expected_q_value = reward + GAMMA * next_q_value
    
    current_model.train()
    results = pc_trainer.train_on_batch(
        inputs=state,
        loss_fn=loss_fn,
        loss_fn_kwargs={
            '_expected_q_value': expected_q_value,
            '_action': action
        },
        is_log_progress=False,
        is_return_results_every_t=False,
        callback_after_t=None,
        is_checking_after_callback_after_t=False,
    )
    
    return results

class DQN(nn.Module):
    def __init__(self, num_inputs, num_actions, predictive_coding=False):
        super(DQN, self).__init__()
        
        self.layers = []
        self.layers.append(nn.Linear(num_inputs, 128))
        if predictive_coding:
            self.layers.append(pc.PCLayer())
        self.layers.append(nn.ReLU())
        self.layers.append(nn.Linear(128, 128))
        if predictive_coding:
            self.layers.append(pc.PCLayer())
        self.layers.append(nn.ReLU())
        self.layers.append(nn.Linear(128, num_actions))
        self.layers = nn.Sequential(*self.layers)
        
    def forward(self, x):
        return self.layers(x)
    
    @torch.no_grad()
    def act(self, state, epsilon):
        self.train(False)
        if random.random() > epsilon:
            state   = torch.FloatTensor(state).unsqueeze(0).to(device)
            q_value = self.forward(state)
            action  = q_value.max(1)[1].data[0].item()
        else:
            action = random.randrange(env.action_space.n)
        return action

def compute_confidence_intervals(matrix):
    upper_confidence_interval = []
    lower_confidence_interval = []

    for iteration in matrix.transpose():
        bootstrapped_means = []
        for B in range(2000):
            bootstrapped_sample = np.random.choice(iteration, len(iteration))
            bootstrapped_means.append(np.mean(bootstrapped_sample))
        bootstrapped_percentiles = np.percentile(np.array(bootstrapped_means), [2.5, 97.5])
        upper_confidence_interval.append(bootstrapped_percentiles[1])
        lower_confidence_interval.append(bootstrapped_percentiles[0])

    return upper_confidence_interval, lower_confidence_interval

def plot_all(bp_reward_matrix, pc_reward_matrix, confidence_intervals=False):
    bp_reward_matrix = np.array(bp_reward_matrix)
    pc_reward_matrix = np.array(pc_reward_matrix)

    # plot BP
    plt.style.use('ggplot')
    bp_average_reward = np.mean(np.mean(bp_reward_matrix, axis=0))
    plt.figure(figsize=(10, 5))
    plt.title('CartPole-v1 : Vanilla DQN with BP (Mean Reward = {:.2f})'.format(bp_average_reward))
    bp_reward_matrix_mean = np.mean(bp_reward_matrix, axis=0)
    plt.plot(bp_reward_matrix_mean, linewidth=2, color='blueviolet', alpha=0.8)

    if confidence_intervals:
        bp_upper_confidence_interval, bp_lower_confidence_interval = compute_confidence_intervals(bp_reward_matrix)

    plt.ylim([0, 500])
    plt.xlabel('episodes')
    plt.ylabel('reward')
    if confidence_intervals:
        plt.fill_between([x for x in range(bp_reward_matrix.shape[1])], bp_upper_confidence_interval, bp_lower_confidence_interval, color='blueviolet', alpha=0.4)
    plt.savefig('./imgs/cartpolev1_bp_rewards_seeds={}_bp_lr={}_bp_optim={}.png'.format(NUM_SEEDS, BP_LR, BP_OPTIM))
    plt.close('all')

    # plot PC
    plt.style.use('ggplot')
    pc_average_reward = np.mean(np.mean(pc_reward_matrix, axis=0))
    plt.figure(figsize=(10, 5))
    plt.title('CartPole-v1 : Vanilla DQN with PC (Mean Reward = {:.2f})'.format(pc_average_reward))
    pc_reward_matrix_mean = np.mean(pc_reward_matrix, axis=0)
    plt.plot(pc_reward_matrix_mean, linewidth=2, color='firebrick', alpha=0.8)

    pc_upper_confidence_interval = []
    pc_lower_confidence_interval = []

    if confidence_intervals:
        pc_upper_confidence_interval, pc_lower_confidence_interval = compute_confidence_intervals(pc_reward_matrix)

    plt.ylim([0, 500])
    plt.xlabel('episodes')
    plt.ylabel('reward')
    if confidence_intervals:
        plt.fill_between([x for x in range(pc_reward_matrix.shape[1])], pc_upper_confidence_interval, pc_lower_confidence_interval, color='firebrick', alpha=0.4)
    plt.savefig('./imgs/cartpolev1_pc_rewards_seeds={}_pc_lr={}_pc_optim={}.png'.format(NUM_SEEDS, PC_LR, PC_OPTIM))
    plt.close('all')

    # plot both on same axis
    plt.style.use('ggplot')
    plt.figure(figsize=(10, 5))
    plt.title('CartPole-v1 : Vanilla DQN')
    plt.plot([x for x in range(bp_reward_matrix.shape[1])], bp_reward_matrix_mean, linewidth=2, color='blueviolet', alpha=0.8, label="BP")
    plt.plot([x for x in range(pc_reward_matrix.shape[1])], pc_reward_matrix_mean, linewidth=2, color='firebrick', alpha=0.8, label="PC")
    plt.legend()

    if confidence_intervals:
        plt.fill_between([x for x in range(bp_reward_matrix.shape[1])], bp_upper_confidence_interval, bp_lower_confidence_interval, color='blueviolet', alpha=0.4)
        plt.fill_between([x for x in range(pc_reward_matrix.shape[1])], pc_upper_confidence_interval, pc_lower_confidence_interval, color='firebrick', alpha=0.4)

    plt.ylim([0, 500])
    plt.xlabel('episodes')
    plt.ylabel('reward')
    plt.savefig('./imgs/cartpolev1_seeds={}_bp_lr={}_bp_optim={}_pc_lr={}_pc_lr={}.png'.format(NUM_SEEDS, BP_LR, BP_OPTIM, PC_LR, PC_OPTIM))
    plt.close('all')

def write_all(bp_reward_matrix, pc_reward_matrix):

    bp_mean_rewards = np.mean(bp_reward_matrix, axis=0)
    pc_mean_rewards = np.mean(pc_reward_matrix, axis=0)

    with open('./data/cartpolev1_seeds={}_bp_lr={}_bp_optim={}_pc_lr={}_pc_lr={}.txt'.format(NUM_SEEDS, BP_LR, BP_OPTIM, PC_LR, PC_OPTIM), 'w') as outfile:
        outfile.write('BP Mean rewards: \n')
        outfile.write(bp_mean_rewards)
        outfile.write('\n')
        outfile.write('PC Mean rewards: \n')
        outfile.write(pc_mean_rewards)
        outfile.write('\n')


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--n-seed", default=NUM_SEEDS)
    parser.add_argument("--bp-lr", default=BP_LR)
    parser.add_argument("--pc-lr", default=PC_LR)
    parser.add_argument("--bp-optim", default=BP_OPTIM)
    parser.add_argument("--pc-optim", default=PC_OPTIM)
    args = parser.parse_args()

    NUM_SEEDS = int(args.n_seed)
    BP_LR = float(args.bp_lr)
    PC_LR = float(args.pc_lr)
    BP_OPTIM = args.bp_optim
    PC_OPTIM = args.pc_optim

    assert torch.cuda.is_available()
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu" )

    # run BP

    bp_loss_matrix = []
    bp_reward_matrix = []

    for seed in range(742, 742+NUM_SEEDS):

        env = gym.make(ENV_ID)

        torch.manual_seed(seed)
        np.random.seed(seed)
        random.seed(seed)
        env.action_space.seed(seed)
        env.seed(seed)

        current_model = DQN(env.observation_space.shape[0], env.action_space.n).to(device)
        target_model  = DQN(env.observation_space.shape[0], env.action_space.n).to(device)
        target_model.eval()
        
        target_model.load_state_dict(current_model.state_dict())

        optimizer = None
        if BP_OPTIM == 'Adam':
            optimizer = optim.Adam
        elif BP_OPTIM == 'SGD':
            optimizer = optim.SGD
        else:
            raise ValueError("Bad argument for optim")

        pc_trainer = pc.PCTrainer(
            current_model,
            T=1,
            update_x_at='all',
            optimizer_x_fn=optim.SGD,
            optimizer_x_kwargs={"lr": 0.1},
            x_lr_discount=0.5,
            update_p_at='last',
            optimizer_p_fn=optimizer,
            optimizer_p_kwargs={"lr": BP_LR},
            plot_progress_at=[],
        )

        replay_buffer = ReplayBuffer(5000)

        losses = []
        all_rewards = []
        episode_reward = 0
        episode = 1
        frame_idx = 0

        state = env.reset()
        while episode <= NUM_EPISODES:
            frame_idx += 1
            epsilon = EPSILON_FINAL + (EPSILON_START - EPSILON_FINAL) * np.exp(-1. * frame_idx / EPSILON_DECAY)
            action = current_model.act(state, epsilon)
            
            next_state, reward, done, _ = env.step(action)
            episode_reward += reward

            if episode_reward < 500:
                replay_buffer.push(state, action, reward, next_state, done)

            state = next_state
            
            if done:
                state = env.reset()
                all_rewards.append(episode_reward)
                episode_reward = 0
                episode += 1
                
            if len(replay_buffer) > BATCH_SIZE:
                loss = compute_td_loss_without_clamp(BATCH_SIZE)['loss'][0]
                losses.append(loss)
                
            if frame_idx % 100 == 0:
                target_model.load_state_dict(current_model.state_dict())

        print("Seed {} complete".format(seed))
        bp_loss_matrix.append(losses)
        bp_reward_matrix.append(all_rewards)
    
    # run PC

    pc_loss_matrix = []
    pc_reward_matrix = []

    for seed in range(742, 742+NUM_SEEDS):

        env = gym.make(ENV_ID)

        torch.manual_seed(seed)
        np.random.seed(seed)
        random.seed(seed)
        env.action_space.seed(seed)
        env.seed(seed)

        current_model = DQN(env.observation_space.shape[0], env.action_space.n, predictive_coding=True).to(device)
        target_model  = DQN(env.observation_space.shape[0], env.action_space.n, predictive_coding=True).to(device)
        target_model.eval()

        replay_buffer = ReplayBuffer(5000)

        update_target(current_model, target_model)

        optimizer = None
        if PC_OPTIM == 'Adam':
            optimizer = optim.Adam
        elif PC_OPTIM == 'SGD':
            optimizer = optim.SGD
        else:
            raise ValueError("Bad argument for optim")

        pc_trainer = pc.PCTrainer(
            current_model,
            T=32,
            update_x_at='all',
            optimizer_x_fn=optim.SGD,
            optimizer_x_kwargs={"lr": 0.05},
            x_lr_discount=0.5,
            update_p_at='last',
            optimizer_p_fn=optimizer,
            optimizer_p_kwargs={"lr": PC_LR},
            plot_progress_at=[],
        )

        losses = []
        all_rewards = []
        episode_reward = 0
        episode = 1
        frame_idx = 0

        state = env.reset()
        while episode <= NUM_EPISODES:
            frame_idx += 1
            epsilon = EPSILON_FINAL + (EPSILON_START - EPSILON_FINAL) * np.exp(-1. * frame_idx / EPSILON_DECAY)
            action = current_model.act(state, epsilon)
            
            next_state, reward, done, _ = env.step(action)
            episode_reward += reward

            if episode_reward < 500:
                replay_buffer.push(state, action, reward, next_state, done)

            state = next_state
            
            if done:
                state = env.reset()
                all_rewards.append(episode_reward)
                episode_reward = 0
                episode += 1
                
            if len(replay_buffer) > BATCH_SIZE:
                results = compute_td_loss_without_clamp(BATCH_SIZE)
                losses.append(results['loss'][-1])
                
            if frame_idx % 100 == 0:
                update_target(current_model, target_model)

        print("Seed {} complete".format(seed))
        pc_loss_matrix.append(losses)
        pc_reward_matrix.append(all_rewards)

    plot_all(bp_reward_matrix, pc_reward_matrix, confidence_intervals=(NUM_SEEDS >= 3))
    write_all(bp_reward_matrix, pc_reward_matrix)
